ElementQueryable

Trait ElementQueryable 

pub trait ElementQueryable {
    // Required method
    fn query(&self, by: By) -> ElementQuery;
}
Trait for enabling the ElementQuery interface.

Required Methods

fn query(&self, by: By) -> ElementQuery

Start an element query using the specified selector.

Dyn Compatibility

This trait is dyn compatible.

In older versions of Rust, dyn compatibility was called "object safety".

Implementations on Foreign Types

Source

impl ElementQueryable for Arc<SessionHandle>

Source

fn query(&self, by: By) -> ElementQuery

Return an ElementQuery instance for more executing powerful element queries.

This uses the builder pattern to construct queries that will return one or more elements, depending on the method specified at the end of the chain.

See ElementQuery for more documentation.

Implementors