pub trait ElementQueryable {
// Required method
fn query(&self, by: By) -> ElementQuery;
}Expand description
Trait for enabling the ElementQuery interface.
Required Methods
Sourcefn query(&self, by: By) -> ElementQuery
fn query(&self, by: By) -> ElementQuery
Start an element query using the specified selector.
Dyn Compatibility
This trait is dyn compatible.
In older versions of Rust, dyn compatibility was called "object safety".
Implementations on Foreign Types
Sourceimpl ElementQueryable for Arc<SessionHandle>
impl ElementQueryable for Arc<SessionHandle>
Sourcefn query(&self, by: By) -> ElementQuery
fn query(&self, by: By) -> ElementQuery
Return an ElementQuery instance for more executing powerful element queries.
This uses the builder pattern to construct queries that will return one or more elements, depending on the method specified at the end of the chain.
See ElementQuery for more documentation.